/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	NJN
 *
 *	$Id: rpi.tcp.h,v 6.1 96/11/22 14:18:07 nevin Rel $
 *
 *	Function:	- TCP client-to-client interface header
 */

#ifndef _RPI_TCP_H
#define _RPI_TCP_H

#include <all_hash.h>
#include <all_list.h>
#include <mpisys.h>
#include <portable.h>
#include <t_types.h>

/*
 * constants
 */
#define INITHASH	((int4) 32)		/* initial hash table size */
#define TCPSHORTMSGLEN	16384			/* max. length of short msg */

/*
 * structures
 *
 * client-to-client envelope
 */
struct c2c_envl {
	int4		ce_len;			/* message length */
	int4		ce_tag;			/* message tag (16 bits) */
	int4		ce_flags;		/* flags */
#define C2CLONG		0x01			/* long protocol */
#define C2CACK		0x02			/* ACK message */
#define C2C2ND		0x04			/* 2nd message chunk */
#define C2CSSEND	0x08			/* synchronous send */
	int4		ce_rank;		/* peer rank */
	int4		ce_cid;			/* message context ID */
	int4		ce_seq;			/* sequence number */
};

/*
 * client-to-client specific process data
 */
struct c2c_proc {
	struct c2c_envl	cp_env;			/* incoming envelope */
	int		cp_sock;		/* socket descriptor */
	int		cp_nenvin;		/* # env. bytes left to read */
	char		*cp_envbuf;		/* envelope buffer pointer */
	int		cp_nmsgin;		/* # msg. bytes left to read */
	char		*cp_msgbuf;		/* message buffer pointer */
	int		cp_extra;		/* # bytes to go into sink */
	MPI_Request	cp_wreq;		/* req writing to process */
	MPI_Request	cp_rreq;		/* reading req */
	MPI_Request	cp_mreq;		/* req to match from */
	int		(*cp_readfn)();		/* read function */
	struct cbuf_msg *cp_bmsg;		/* buffering? */
	int		cp_nbfde;		/* # buffered envelopes */
};

/*
 * client-to-client specific request data
 */
struct c2c_req {				
	int		cq_state;		/* state of request */
#define	C2CDONE		1			/* all data transfer done */
#define C2CWRITE	2			/* writing to socket */
#define C2CREAD		3			/* reading from socket */
#define C2CSENDSELF	4			/* sending to self */
	int4		cq_peer;		/* matched peer */
	struct c2c_envl cq_env;			/* envelope (synch + state) */	
	struct c2c_envl cq_outenv;		/* byteordered outgoing env. */
	int		cq_nenvout;		/* # env. bytes left to write */
	int		cq_nmsgout;		/* # msg. bytes left to write */
	char		*cq_envbuf;		/* envelope buffer pointer*/
	char		*cq_msgbuf;		/* message buffer pointer */
	int		(*cq_adv)();		/* advance function */
};

/*
 * CID hash table entry
 */
struct cbuf_cid {				
	int4		cb_cid;			/* context ID */
	MPI_Comm	cb_comm;		/* communicator */
	LIST		*cb_envs;		/* envelope list */
};

/*
 * message/envelope entry
 */
struct cbuf_msg {
	struct c2c_proc *cm_proc;		/* source process */
	struct c2c_envl	cm_env;			/* envelope */
	char		*cm_buf;		/* short message data */
	MPI_Request	cm_req;			/* send req if sender == self */
};

/*
 * internal macros
 */
#define commdead_m(r)	\
	(((r)->rq_comm->c_flags & (LAM_CLDEAD | LAM_CRDEAD))	\
	? _c2c_comm_dead(r) : 0)

/*
 * data conversion for client-to-client envelopes
 */
#if WORDS_BIGENDIAN

#define tcp_set_out_envelope_m(r) \
	{ \
		(r).cq_nenvout = sizeof(struct c2c_envl); \
		(r).cq_envbuf = (char *) &((r).cq_env); \
	}

#else	/* not WORDS_BIGENDIAN */

#define tcp_set_out_envelope_m(r) \
	if (lam_homog) { \
		(r).cq_nenvout = sizeof(struct c2c_envl); \
		(r).cq_envbuf = (char *) &((r).cq_env); \
	} \
	else { \
		(r).cq_nenvout = sizeof(struct c2c_envl); \
		(r).cq_outenv = (r).cq_env; \
		mltoti4((int4 *) &((r).cq_outenv), \
				sizeof(struct c2c_envl) / sizeof(int4)); \
		(r).cq_envbuf = (char *) &((r).cq_outenv); \
	}

#endif	/* not WORDS_BIGENDIAN */

/*
 * prototypes of LAM MPI library client-to-client internal functions
 */
#ifndef __ARGS
#if __STDC__ || defined(c_plusplus) || defined(__cplusplus)
#define __ARGS(a)	a
#else
#define __ARGS(a)	()
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int	_c2c_comm_dead  __ARGS((MPI_Request));
extern int	_c2c_envl_cmp __ARGS((struct c2c_envl *, struct c2c_envl *));
extern void	_c2c_fill_mpi_status __ARGS((MPI_Request, int, int, int));
extern void	_c2c_fill_wildcards __ARGS((MPI_Request, struct c2c_envl *));

extern int	_cbuf_init __ARGS((void));
extern int	_cbuf_append __ARGS((struct cbuf_msg *));
extern void	_cbuf_delete __ARGS((struct cbuf_msg *));
extern void	_cbuf_end __ARGS((void));
extern struct cbuf_msg
		*_cbuf_find __ARGS((struct c2c_envl *));

extern int	_tcp_advmultiple __ARGS((void));
extern int	_tcp_adv1 __ARGS((void));
extern int	_tcp_proc_read_env __ARGS((struct c2c_proc *));
extern int	_tcp_req_write_long __ARGS((struct c2c_proc *, MPI_Request));
extern int	_tcp_req_write_short __ARGS((struct c2c_proc *, MPI_Request));
extern int	_tcp_req_write_synch __ARGS((struct c2c_proc *, MPI_Request));
extern int	_tcp_req_read __ARGS((struct c2c_proc *, MPI_Request));
extern int	_tcp_req_probe __ARGS((struct c2c_proc *, MPI_Request));
extern int	_tcp_buffered_adv __ARGS((MPI_Request, struct cbuf_msg *));

#ifdef __cplusplus
}
#endif

#endif	/* _RPI_TCP_H */
